/* newfiles.c
 * Copyright (C) 1991 by Chin Huang
 *
 * Eliminate repeated entries in PCBoard file list.
 * Compile with:
 *	tcc -mc newfiles.c
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static char *progname = "newfiles";

/* maximum path length */
#define MAXPATH 80

/* maximum input line length */
#define MAXLINELEN 81

/* maximum symbol length */
#define MAXKEYLEN 22

/* maximum number of conferences */
#define MAX_CONF 20

/* maximum conference name length */
#define MAX_CONF_NAME 12

/* default conference name */
#define DEFAULT_CONFERENCE "MAIN"

/* number of symbols in symbol block */
#define SYM_BLOCK_LEN 32

/* Symbols have a name. */
typedef struct {
	char name[MAXKEYLEN];
} Symbol;

/* symbol table block */
typedef struct _symbol_block {
	struct _symbol_block *next;		/* next symbol in list */
	int num_used;				/* number of entries used */
	Symbol symbol[SYM_BLOCK_LEN];		/* symbols */
} SymbolBlock;

/* hash table length */
#define SYM_MAX_HASH 256

/* symbol table */
typedef struct {
	SymbolBlock *table[SYM_MAX_HASH];	/* hash buckets */
} SymbolTable;

/* conference data */
typedef struct {
	char name[MAX_CONF_NAME];
	FILE *fp;
} Conference;

Conference conference[MAX_CONF];	/* table of conferences */
Conference *cur_conf;			/* current conference */
int num_conf;				/* number of conferences */
char data_dir[MAXPATH];			/* data directory */
SymbolTable symbols;			/* files already encountered */
long high_limit = 196608L;		/* file size high limit */
long low_limit = 131072L;		/* flle size low limit */

/* Initialize symbol table.
 */
void
init_symbol_table (st)
SymbolTable *st;
{
    int i;

    for (i = 0; i < SYM_MAX_HASH; ++i)
	st->table[i] = NULL;
}

/* Simple hash function mapping a symbol name to a hash bucket.
 */
static int
hash (name)
char *name;
{
    return (name[0] + name[1] + name[2]) % SYM_MAX_HASH;
}

/* Search the list of symbols <list> for the symbol <name>.
 * Return a pointer to the symbol if found or NULL if not found.
 */
static Symbol *
find_symbol_list (list, name)
SymbolBlock *list;
char *name;
{
    SymbolBlock *sym;
    int i;

    for (sym = list; sym != NULL; sym = sym->next) {
        for (i = 0; i < sym->num_used; ++i) {
	    if (strcmp(sym->symbol[i].name, name) == 0)
		return &(sym->symbol[i]);
	}
    }
    return NULL;
}

/* Look for symbol <name> in symbol table.
 * Return 1 if found or 0 if not found.
 */
Symbol *
find_symbol (st, name)
SymbolTable *st;
char *name;
{
    return find_symbol_list(st->table[hash(name)], name);
}

/* Allocate new symbol table block.
 */
SymbolBlock *
new_symbol_block (void)
{
    SymbolBlock *sym;

    if ((sym = (SymbolBlock *)malloc(sizeof(SymbolBlock))) == NULL) {
	fprintf(stderr, "%s: out of memory allocating symbol\n", progname);
	exit(1);
    }
    sym->num_used = 0;
    return sym;
}

/* Add the symbol to the symbol table.  Return a pointer to the new symbol.
 */
Symbol *
add_symbol (st, name)
SymbolTable *st;
char *name;
{
    SymbolBlock *sb;
    Symbol *sym;
    int h;

    h = hash(name);
    if ((sb = st->table[h]) == NULL || sb->num_used == SYM_BLOCK_LEN) {
	sb = new_symbol_block();
	sb->next = st->table[h];
	st->table[h] = sb;
    }
    sym = &(sb->symbol[sb->num_used++]);
    strcpy(sym->name, name);
    return sym;
}

/* Extract key string consisting of file name and size from input line.
 */
void
get_key (line, key)
char *line, *key;
{
    strncpy(key, line, MAXKEYLEN-1);
    key[MAXKEYLEN-1] = '\0';
}

/* Set the current conference.
 */
void
set_conference (name)
char *name;		/* conference name */
{
    int i;
    char conf_name[MAX_CONF_NAME], path[MAXPATH];
    char line[MAXLINELEN], key[MAXKEYLEN];
    char *stat;

    strcpy(conf_name, name);
    strupr(conf_name);
    for (i = 0; i < num_conf; ++i) {
	if (strcmp(conference[i].name, conf_name) == 0) {
	    cur_conf = &conference[i];
	    return;
	}
    }

    if (num_conf >= MAX_CONF) {
	fprintf(stderr, "%s: too many conferences\n", progname);
	exit(1);
    }
    cur_conf = &conference[num_conf++];
    strcpy(cur_conf->name, conf_name);

    sprintf(path, "%s\\%s.NEW", data_dir, conf_name);
    if ((cur_conf->fp = fopen(path, "a+")) == NULL) {
	fprintf(stderr, "%s: cannot open %s\n", progname, path);
	exit(1);
    }
    fseek(cur_conf->fp, 0L, SEEK_SET);

    while (fgets(line, MAXLINELEN, cur_conf->fp) != NULL) {
        if (line[0] != ' ') {
	    get_key(line, key);
	    if (find_symbol(&symbols, key) == NULL) {
		add_symbol(&symbols, key);
	    }
	}
    }
}

/* Read through file list until file name found.  If found, then return 1
 * and store first line in line.
 */
int
skip_to_file_entry (inf, line)
FILE *inf;
char *line;
{
    int conf_num;
    char conf_name[MAX_CONF_NAME];

    do {
	if (strncmp(line, "New files for", 13) == 0) {
	    fputs(line, stdout);
	    continue;
	}
	if (strncmp(line, "\304\304\304", 3) == 0) {
	    fputs(line, stdout);
	    continue;
	}
	if (strncmp(line, "Conference", 10) == 0) {
	    fputs(line, stdout);
	    sscanf(line, "Conference %d %s", &conf_num, conf_name);
	    set_conference(conf_name);
	    continue;
	}
	if (line[0] != ' ' && strlen(line) >= MAXKEYLEN-1 &&
	    strchr(line, ' ') != NULL) {
	    return 1;
	}
	fputs(line, stdout);
    } while (fgets(line, MAXLINELEN, inf) != NULL);
    return 0;
}

/* Copy list of files to standard output, eliminating entries that have
 * already been encountered.
 */
void
copy_list (path)
char *path;
{
    FILE *infp;
    char line[MAXLINELEN], key[MAXKEYLEN];
    char *stat;

    if ((infp = fopen(path, "r")) == NULL) {
	fprintf(stderr, "%s: cannot open file %s\n", progname, path);
	return;
    }
    stat = fgets(line, MAXLINELEN, infp);
    while (stat != NULL && skip_to_file_entry(infp, line)) {
	get_key(line, key);
	if (find_symbol(&symbols, key) == NULL) {
	    add_symbol(&symbols, key);
	    fputs(line, stdout);
	    fputs(line, cur_conf->fp);
	    while ((stat = fgets(line, MAXLINELEN, infp)) != NULL) {
		if (line[0] != ' ')
		    break;
		fputs(line, stdout);
		fputs(line, cur_conf->fp);
	    }
	} else {
	    while ((stat = fgets(line, MAXLINELEN, infp)) != NULL) {
		if (line[0] != ' ')
		    break;
	    }
	}
    }
    fclose(infp);
}

#define BUFSIZE 2048

/* Trim the size of the file lists by discarding all entries before the first
 * duplicated entry.
 */
void
close_conferences (void)
{
    int i;
    size_t n;
    long offset;
    FILE *outfp;
    char buf[BUFSIZE], tmppath[MAXPATH], confpath[MAXPATH];

    for (i = 0; i < num_conf; ++i) {
        cur_conf = &conference[i];
	if ((offset = ftell(cur_conf->fp)) > high_limit) {

	    sprintf(tmppath, "%s\\NEWFILES.TMP", data_dir);
	    if ((outfp = fopen(tmppath, "w")) == NULL) {
		fprintf(stderr, "%s: cannot open %s\n", progname, tmppath);
		exit(1);
	    }

	    fseek(cur_conf->fp, offset-low_limit, SEEK_SET);
	    fgets(buf, BUFSIZE, cur_conf->fp);

	    while ((n = fread(buf, 1, BUFSIZE, cur_conf->fp)) > 0) {
		fwrite(buf, 1, n, outfp);
	    }
	    fclose(outfp);
	    fclose(cur_conf->fp);

	    sprintf(confpath, "%s\\%s.NEW", data_dir, cur_conf->name);
	    remove(confpath);
	    rename(tmppath, confpath);
	} else {
	    fclose(cur_conf->fp);
	}
    }
}

void
main (argc, argv)
int argc;
char **argv;
{
    int i;
    char path[64], *s;

    if (argc < 2) {
        fprintf(stderr, "%s version 1.0, Copyright (C) 1991 by Chin Huang\n",
	    progname);
        fprintf(stderr, "Removes duplicate entries from PCBoard file lists\n");
        fprintf(stderr, "and produces new file list on standard output.\n");
	fprintf(stderr, "usage: %s listfile ...\n", progname);
	exit(1);
    }

    if ((s = getenv("NEWFILES")) != NULL) {
	strcpy(data_dir, s);
    } else {
	strcpy(data_dir, ".");
    }
    if ((i = strlen(data_dir)) > 0 && data_dir[i-1] == '\\')
	data_dir[i-1] = '\0';

    init_symbol_table(&symbols);
    num_conf = 0;
    set_conference(DEFAULT_CONFERENCE);

    for (i = 1; i < argc; ++i) {
	copy_list(argv[i]);
    }
    close_conferences();
}
